//
// pp.c
// 
// Device driver de caracteres (Kernel 2.4) para puerta paralelo.
// Demostracin para el curso de Postgrado 'Temas
// de Instrumentacin Electrnica'.
// 
// OJO: No olvidarse de crear el nodo con:
// 	mknod /dev/pp0 c PP_MAJOR 0
// y de instalar el mdulo
// 
// Alejandro Veiga - Septiembre 2002
//

#define MODULE
// Kernel includes
// Standard in kernel modules
#include <linux/kernel.h>   	// We're doing kernel work
#include <linux/module.h>   	// Specifically, a module
// For putting processes to sleep and waking them up
#include <linux/sched.h>
#include <linux/wrapper.h>
#include <asm/io.h>		// inb and outb
#include <asm/uaccess.h>  	// for get_user and put_user
#include "pp.h"

int pp_busy = LIBRE;
static struct wait_queue *pp_wait_q;

// IRQ /////////////////////////////////////////////////
int pp_irq(int irq)
{
	module_wake_up(&pp_wait_q);
	printk("Kernel printk: interrupcin %d\n", irq);
}


// READ ////////////////////////////////////////////////
ssize_t pp_read(struct file *f, char *buf, size_t count, loff_t *off)
{
	unsigned int data;
	module_interruptible_sleep_on(&pp_wait_q);
	data = inb(PP_DATA);
	put_user(data, buf);
	// printk("Kernel printk: El driver leyo: %d\n", data);
	return 0;
}

// WRITE ///////////////////////////////////////////////
ssize_t pp_write(struct file *f, const char *buf, size_t count, loff_t *off)
{
        char c;
	const char *temp;

	temp=buf;
        while(count>0){
                get_user(c,temp); /* Get next char */
                outb(c|0x01,PP_DATA); /* Send it to the PP */
                count--;
                temp++;
        }
	return temp-buf;
}

// OPEN ////////////////////////////////////////////////
int pp_open(struct inode *ino, struct file *f)
{
	unsigned int minor = MINOR(ino->i_rdev);
	int ret_code;
	if (minor != 0) return -ENODEV;
	if (pp_busy ==  OCUPADO) return -EBUSY;
        pp_busy = OCUPADO;
	ret_code = request_irq(PP_IRQ,(void *)pp_irq,SA_INTERRUPT,"pp",NULL);
	if (ret_code) {
		printk("Kernel printk: pp_open error init irq %d\n", PP_IRQ);
		return ret_code;
	} else {
		printk("Kernel printk: pp_open OK\n");
	}	

	/////////////////////////////////////////////////////////
	// En Linux no es necesario hacer los siguientes pasos	
	// Port 0x21 -> interrupt mask (write 0 to bit 7 to
	// enable irq7 - Do not modify the others!!)
	//outb(inb(0x21)&(~0x80),0x21);
	// Tell the PC the irq has been processed
	//outb(0x20,0x20);
	/////////////////////////////////////////////////////////

	// Set PP_DATA.0 (to get Vcc for the pullup resistor)
	outb(inb(PP_DATA)|0x01,PP_DATA);
	// Enable PP IRQ seting PP_CONTROL.4
	outb(inb(PP_CONTROL)|0x10,PP_CONTROL);
	printk("Kernel printk: pp irq OK\n");
	sti();
  	return 0;
}

// RELEASE /////////////////////////////////////////////
int pp_release(struct inode *ino, struct file *f)
{
	cli();
	// Port 0x21 -> interrupt mask (write 1 to bit 7 to
	// disable irq7 - Do not modify the others!!)
	outb(inb(0x21)|(0x80),0x21);
	// Disable PP IRQ
	outb(inb(PP_CONTROL)&(~0x10),PP_CONTROL);
	printk("Kernel printk: pp irq disabled\n");
	// Free it
	free_irq(PP_IRQ,NULL);
	pp_busy  = LIBRE;
	printk("Kernel printk: pp_release OK\n");
}

// FOPS //////////////
struct file_operations pp_fops = {
	read:		pp_read,
	write:		pp_write,
	open:		pp_open,
	release:	pp_release
};

// INIT MODULE /////////////////////////////////////////
int init_module(void) 
{
	if (register_chrdev(PP_MAJOR, "pp", &pp_fops)) {
	  printk("Kernel printk: pp register_chrdev failed\n");
	  return -EIO;
	} 
        return 0;
}

// CLEANUP MODULE //////////////////////////////////////
int cleanup_module(void) 
{
	if (pp_busy) 
		printk("Kernel printk: pp busy\n");
	if (unregister_chrdev(PP_MAJOR, "pp") != 0) {
		printk("Kernel printk pp cleanup_module failed\n");
	} 
	else {
	}
}
